use rspotify::model::PlayableItem::{Episode, Track};
use rspotify::model::{AlbumId, PlaylistId, TrackId};
use rspotify::prelude::*;
use rspotify::{ClientCredsSpotify, Credentials};
use std::sync::Arc;
use std::time::Duration;

#[derive(Debug, Clone, PartialEq, Eq, Hash)]
pub enum ContentKind {
    Track(String),
    Album(String),
    Playlist(String),
    Podcast(String),
    Episode(String),
}

#[derive(Debug, Clone, PartialEq, Eq, Hash)]
pub enum PlayableKind {
    Track(TrackInfo),
    Podcast(EpisodeInfo),
}

#[derive(Debug, Clone, PartialEq, Eq, Hash)]
pub struct TrackInfo {
    pub(crate) name: String,
    pub(crate) artists: Vec<String>,
    pub(crate) duration: Duration,
}

#[derive(Debug, Clone, PartialEq, Eq, Hash)]
pub struct EpisodeInfo {
    pub(crate) name: String,
    pub(crate) show: String,
    pub(crate) duration: Duration,
    pub(crate) description: String,
    pub(crate) languages: Vec<String>,
    pub(crate) release_date: String,
}

#[derive(Debug, Clone, PartialEq, Eq, Hash)]
pub struct AlbumInfo {
    pub(crate) name: String,
    pub(crate) artists: Vec<String>,
    pub(crate) genres: Vec<String>,
    pub(crate) tracks: Vec<TrackInfo>,
}

#[derive(Debug, Clone, PartialEq, Eq, Hash)]
pub struct PlaylistInfo {
    pub(crate) name: String,
    pub(crate) artists: Vec<String>,
    pub(crate) tracks: Vec<PlayableKind>,
    pub(crate) owner: Option<String>,
}

#[derive(Clone, Debug)]
pub(crate) struct Client {
    client: Arc<ClientCredsSpotify>,
}

impl Client {
    pub(crate) async fn new() -> Self {
        let spotify_creds = Credentials::from_env()
            .expect("RSPOTIFY_CLIENT_ID and RSPOTIFY_CLIENT_SECRET not found.");
        let mut spotify = ClientCredsSpotify::new(spotify_creds);
        spotify.request_token().await.unwrap();
        Client {
            client: Arc::new(spotify),
        }
    }

    pub(crate) fn new_with_dependencies(client: ClientCredsSpotify) -> Self {
        Client {
            client: Arc::new(client),
        }
    }

    pub async fn get_track(&self, id: &str) -> Option<TrackInfo> {
        // FIXME should we really return Option here? We're hiding a possible error or a entry not found
        let track_id = match TrackId::from_id(id) {
            Ok(track) => track,
            Err(_e) => return None,
        };

        match self.client.track(&track_id).await {
            Ok(track) => Some(TrackInfo {
                name: track.name,
                artists: track.artists.iter().map(|x| x.name.clone()).collect(),
                duration: track.duration,
            }),
            Err(_e) => return None,
        }
    }

    pub async fn get_album(&self, id: &str) -> Option<AlbumInfo> {
        let album_id = match AlbumId::from_id(id) {
            Ok(album) => album,
            Err(_e) => return None,
        };

        match self.client.album(&album_id).await {
            Ok(album) => Some(AlbumInfo {
                name: album.name,
                artists: album.artists.iter().map(|x| x.name.clone()).collect(),
                genres: album.genres,
                tracks: album
                    .tracks
                    .items
                    .iter()
                    .map(|t| TrackInfo {
                        name: t.name.clone(),
                        artists: t.artists.iter().map(|x| x.name.clone()).collect(),
                        duration: t.duration,
                    })
                    .collect(),
            }),
            Err(_e) => None,
        }
    }

    pub async fn get_playlist(&self, id: &str) -> Option<PlaylistInfo> {
        let playlist_id = match PlaylistId::from_id(id) {
            Ok(playlist) => playlist,
            Err(_e) => return None,
        };

        match self.client.playlist(&playlist_id, None, None).await {
            Ok(playlist) => Some(PlaylistInfo {
                name: playlist.name,
                artists: playlist
                    .tracks
                    .items
                    .iter()
                    .flat_map(|p| {
                        match &p.track {
                            Some(t) => match t {
                                Track(t) => t.artists.iter().map(|a| a.name.clone()).collect(),
                                Episode(e) => vec![e.show.publisher.clone()],
                            },
                            None => Vec::new(),
                        }
                        .into_iter()
                    })
                    .collect(),
                tracks: playlist
                    .tracks
                    .items
                    .iter()
                    .map(|p| match &p.track {
                        Some(t) => match t {
                            Track(t) => Some(PlayableKind::Track(TrackInfo {
                                name: t.name.clone(),
                                artists: t.artists.iter().map(|a| a.name.clone()).collect(),
                                duration: t.duration,
                            })),
                            Episode(e) => Some(PlayableKind::Podcast(EpisodeInfo {
                                name: e.name.clone(),
                                show: e.show.name.clone(),
                                duration: e.duration,
                                description: e.description.clone(),
                                languages: e.languages.clone(),
                                release_date: e.release_date.clone(),
                            })),
                        },
                        None => None,
                    })
                    .filter(|i| i.is_some())
                    .map(|i| i.unwrap())
                    .collect(),
                owner: playlist.owner.display_name,
            }),
            Err(_e) => None,
        }
    }
}

fn get_id_in_url(url: &str) -> Option<&str> {
    url.rsplit('/')
        .next()
        .and_then(|x| x.split(' ').next())
        .and_then(|x| x.split('?').next())
}

fn get_id_in_uri(uri: &str) -> Option<&str> {
    uri.rsplit(':').next()
}

pub fn get_entry_kind(uri: &str) -> Option<ContentKind> {
    if uri.contains("spotify:track:") {
        let track_id = get_id_in_uri(uri);
        return match track_id {
            Some(id) => Some(ContentKind::Track(id.to_string())),
            None => None,
        };
    }
    if uri.contains("https://open.spotify.com/track/") {
        let track_id = get_id_in_url(uri);
        return match track_id {
            Some(id) => Some(ContentKind::Track(id.to_string())),
            None => None,
        };
    }
    if uri.contains("spotify:album:") {
        let track_id = get_id_in_uri(uri);
        return match track_id {
            Some(id) => Some(ContentKind::Album(id.to_string())),
            None => None,
        };
    }
    if uri.contains("https://open.spotify.com/album/") {
        let album_id = get_id_in_url(uri);
        return match album_id {
            Some(id) => Some(ContentKind::Album(id.to_string())),
            None => None,
        };
    }
    if uri.contains("spotify:playlist:") {
        let track_id = get_id_in_uri(uri);
        return match track_id {
            Some(id) => Some(ContentKind::Album(id.to_string())),
            None => None,
        };
    }
    if uri.contains("https://open.spotify.com/playlist/") {
        let playlist_id = get_id_in_url(uri);
        return match playlist_id {
            Some(id) => Some(ContentKind::Playlist(id.to_string())),
            None => None,
        };
    }
    if uri.contains("spotify:show:") {
        let track_id = get_id_in_uri(uri);
        return match track_id {
            Some(id) => Some(ContentKind::Album(id.to_string())),
            None => None,
        };
    }
    if uri.contains("https://open.spotify.com/show/") {
        let playlist_id = get_id_in_url(uri);
        return match playlist_id {
            Some(id) => Some(ContentKind::Podcast(id.to_string())),
            None => None,
        };
    }
    if uri.contains("spotify:episode:") {
        let track_id = get_id_in_uri(uri);
        return match track_id {
            Some(id) => Some(ContentKind::Album(id.to_string())),
            None => None,
        };
    }
    if uri.contains("https://open.spotify.com/episode/") {
        let playlist_id = get_id_in_url(uri);
        return match playlist_id {
            Some(id) => Some(ContentKind::Episode(id.to_string())),
            None => None,
        };
    }
    return None;
}
