use std::error::Error;
use std::sync::Arc;

#[derive(Debug, Clone, PartialEq, Eq, Hash)]
pub(crate) struct VideoSearch {
    pub(crate) items: Vec<Video>,
}

#[derive(Debug, Clone, PartialEq, Eq, Hash)]
pub(crate) struct Video {
    pub(crate) title: String,
    pub(crate) video_id: String,
    pub(crate) author: String,
    pub(crate) author_id: String,
    pub(crate) author_url: String,
    pub(crate) length_seconds: u64,
    pub(crate) description: String,
    pub(crate) description_html: String,
    pub(crate) view_count: u64,
    pub(crate) published: u64,
    pub(crate) published_text: String,
    pub(crate) live_now: bool,
    pub(crate) paid: bool,
    pub(crate) premium: bool,
}

type SearchSortBy = str;

const BY_RELEVANCE: &SearchSortBy = "relevance";
const BY_RATING: &SearchSortBy = "rating";
const BY_UPLOAD_DATE: &SearchSortBy = "upload_date";
const BY_VIEW_COUNT: &SearchSortBy = "view_count";

#[derive(Debug, Clone)]
pub(crate) struct Client {
    client: Arc<invidious::asynchronous::Client>,
}

impl Client {
    pub(crate) async fn new() -> Self {
        // TODO check for a stable instance
        let client =
            invidious::asynchronous::Client::new(String::from("https://inv.bp.projectsegfau.lt"));

        Client {
            client: Arc::new(client),
        }
    }

    pub(crate) fn new_with_dependencies(client: invidious::asynchronous::Client) -> Self {
        Client {
            client: Arc::new(client),
        }
    }

    pub async fn search_video(
        &self,
        keyword: &str,
        sort_by: Option<&SearchSortBy>,
    ) -> Result<VideoSearch, Box<dyn Error>> {
        let mut query = Vec::<String>::new();
        query.push(format!("{}={}", "q", keyword));
        match sort_by {
            Some(sorting_type) => query.push(format!("{}={}", "sort_by", sorting_type)),
            None => {}
        }
        query.push(format!("{}={}", "type", "video"));

        let generated_query = query.join(",");
        let search = self.client.search(Some(generated_query.as_str())).await?;

        let videos: Vec<Video> = search
            .items
            .iter()
            .map(|search_result| match search_result {
                invidious::structs::hidden::SearchItem::Video {
                    title,
                    videoId,
                    author,
                    authorId,
                    authorUrl,
                    lengthSeconds,
                    videoThumbnails: _,
                    description,
                    descriptionHtml,
                    viewCount,
                    published,
                    publishedText,
                    liveNow,
                    paid,
                    premium,
                } => Some(Video {
                    title: title.clone(),
                    video_id: videoId.clone(),
                    author: author.clone(),
                    author_id: authorId.clone(),
                    author_url: authorUrl.clone(),
                    length_seconds: lengthSeconds.clone(),
                    description: description.clone(),
                    description_html: descriptionHtml.clone(),
                    view_count: viewCount.clone(),
                    published: published.clone(),
                    published_text: publishedText.clone(),
                    live_now: liveNow.clone(),
                    paid: paid.clone(),
                    premium: premium.clone(),
                }),
                _ => None,
            })
            .filter(|x| x.is_some())
            .map(|x| x.unwrap())
            .filter(|x| !x.premium && !x.paid && !x.live_now)
            .collect();

        Result::Ok(VideoSearch { items: videos })
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn should_search_properly() {
        let client = tokio_test::block_on(Client::new());

        let result = tokio_test::block_on(client.search_video(
            "vfdskvnfdsjklvnfdsjklvnfsdjkldsvmdlfmvkdfslvsdfmklsdvlvfdnjkvnfdsjkvnfsdjk",
            Some(BY_UPLOAD_DATE),
        ))
        .unwrap();
        assert_eq!(30, result.items.len())
    }
}
